/* packcd: pack files onto fixed-size media. (I'd worry about the "cd" in the
   name, except that "tar" appears to have survived quite happily for 30 years
   despite the death of tapes...)
   Adam Sampson <azz@gnu.org>
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

typedef unsigned long long size_type;
#define SIZE_FORMAT "%lld"

enum pack_mode {
	MODE_SEQUENTIAL,
	MODE_SORT_FIRST_FIT
};

enum command_mode {
	CMODE_MKISOFS,
	CMODE_BURN,
	CMODE_MV
};

enum pack_mode mode = MODE_SEQUENTIAL;
enum command_mode cmode = CMODE_BURN;
size_type size = 700 * 1024 * 1024;
int quiet_mode = 0;

/* These are rather useful, and should probably be moved elsewhere. */
#define DECLARE_VECTOR(name, type) \
	type *name; int name##_size; int name##_capacity
#define INIT_VECTOR(name) \
	do { \
		name = NULL; \
		name##_capacity = name##_size = 0; \
	} while (0)
#define FREE_VECTOR(name) \
	do { \
		if (name != NULL) \
			free(name); \
	} while (0)
#define VECTOR_SIZE(name) \
	name##_size
#define RESIZE_VECTOR(name, size) \
	do { \
		name##_size = (size); \
		if (name##_size > name##_capacity) { \
			name##_capacity = name##_size; \
			name = realloc(name, name##_capacity * sizeof *name); \
			if (name == NULL) \
				die("out of memory", NULL); \
		} \
	} while (0)
#define APPEND_VECTOR(name, val) \
	do { \
		RESIZE_VECTOR(name, name##_size + 1); \
		name[name##_size - 1] = val; \
	} while (0)

struct file {
	const char *name;
	size_type size;
};

struct disc {
	DECLARE_VECTOR(files, struct file *);
	size_type count;
};

DECLARE_VECTOR(files, struct file);
DECLARE_VECTOR(discs, struct disc);

void die(const char *msg, const char *arg) {
	fprintf(stderr, "%s%s%s\n",
		msg, arg == NULL ? "" : ": ", arg == NULL ? "" : arg);
	exit(20);
}

void usage() {
	die("usage: packcd [-cfibmq] [-s SIZE] FILES ...", NULL);
}

int add_disc() {
	int n = VECTOR_SIZE(discs);
	RESIZE_VECTOR(discs, n + 1);

	INIT_VECTOR(discs[n].files);
	discs[n].count = 0;

	return n;
}

int size_compare(const void *a_, const void *b_) {
	const struct file *a = (const struct file *) a_;
	const struct file *b = (const struct file *) b_;

	if (a->size < b->size)
		return 1;
	else if (a->size > b->size)
		return -1;
	else
		return 0;
}

int main(int argc, char **argv) {
	int nfiles, i, j;
	size_type total;

	while (1) {
		char c = getopt(argc, argv, "+cfibms:q");
		if (c == -1)
			break;
		switch (c) {
		case 'c':
			mode = MODE_SEQUENTIAL;
			break;
		case 'f':
			mode = MODE_SORT_FIRST_FIT;
			break;
		case 'i':
			cmode = CMODE_MKISOFS;
			break;
		case 'b':
			cmode = CMODE_BURN;
			break;
		case 'm':
			cmode = CMODE_MV;
			break;
		case 'q':
			quiet_mode = 1;
			break;
		case 's':
			size = atol(optarg);
			if (strchr(optarg, 'k') != NULL)
				size *= 1024;
			if (strchr(optarg, 'm') != NULL)
				size *= 1024 * 1024;
			break;
		default:
			usage();
		}
	}

	nfiles = argc - optind;
	if (nfiles == 0)
		usage();
	INIT_VECTOR(files);
	RESIZE_VECTOR(files, nfiles);
	INIT_VECTOR(discs);

	for (i = 0; i < nfiles; i++) {
		struct stat st;
	
		files[i].name = argv[optind + i];
		if (stat(files[i].name, &st) < 0)
			die("unable to stat", files[i].name);
		files[i].size = st.st_size;
		if (files[i].size > size)
			die("file too large for media", files[i].name);
	}

	if (!quiet_mode) {
		printf("# packcd: packing to media of size "
		       SIZE_FORMAT "\n\n", size);
	}

	if (mode == MODE_SEQUENTIAL) {
		j = add_disc();
		for (i = 0; i < VECTOR_SIZE(files); i++) {
			if (discs[j].count + files[i].size > size)
				j = add_disc();
			APPEND_VECTOR(discs[j].files, &files[i]);
			discs[j].count += files[i].size;
		}
	} else if (mode == MODE_SORT_FIRST_FIT) {
		qsort(files, VECTOR_SIZE(files), sizeof *files, size_compare);
		for (i = 0; i < VECTOR_SIZE(files); i++) {
			for (j = 0; j < VECTOR_SIZE(discs); j++) {
				if (discs[j].count + files[i].size <= size)
					goto found_disc;
			}
			j = add_disc();
			found_disc:
			APPEND_VECTOR(discs[j].files, &files[i]);
			discs[j].count += files[i].size;
		}
	} else {
		die("unknown or unimplemented mode", NULL);
	}

	total = 0;
	for (i = 0; i < VECTOR_SIZE(discs); i++) {
		total += discs[i].count;
		if (!quiet_mode) {
			printf("# disc %d, used " SIZE_FORMAT
				" of " SIZE_FORMAT " bytes (%.1f%%):\n",
				i, discs[i].count, size,
				(100.0 * discs[i].count) / size);
		}
		if (cmode == CMODE_MKISOFS)
			printf("mkisofs -R -J -o /tmp/tmp.iso");
		else if (cmode == CMODE_BURN)
			printf("burn");
		else if (cmode == CMODE_MV)
			printf("mkdir -p %03d\nmv", i);
		for (j = 0; j < VECTOR_SIZE(discs[i].files); j++) {
			printf(" \"%s\"", discs[i].files[j]->name);
		}
		if (cmode == CMODE_MV)
			printf(" %03d/", i);
		printf("\n");
		if (!quiet_mode)
			printf("\n");
	}
	if (!quiet_mode) {
		printf("# total: used " SIZE_FORMAT
			" of " SIZE_FORMAT " bytes (%.1f%%)\n",
			total, size * VECTOR_SIZE(discs),
			(100.0 * total) / (size * VECTOR_SIZE(discs)));
	}

	return 0;
}

